import requests
import json
from tcrpy.models import *
from tcrpy.enums import *
from tcrpy.auth import auth
import datetime as dt
from tcrpy.utils import millisecond_stamp_to_datetime

class GetGridData:
    def __init__(self, headers):
        self.headers = headers

class api:
    baseUrl = "http://apps.tcrsoftware.com/tcr_2"
    getGridDataURL = baseUrl + "/webservices/data.asmx/GetGridData"
    getTicketURL = baseUrl + "/webservices/Tickets.asmx/GetTicket"
    getJobURL = baseUrl + "/webservices/Jobs.asmx/GetJobByID"
    getCustomerURL = baseUrl + "/webservices/Customers.asmx/GetCustomer"
    getGridURL = baseUrl + "/webservices/config.asmx/GetGrid"
    getGridByIDURL = baseUrl + "/webservices/config.asmx/GetGridByID"
    getCompanyURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetCompany"
    getUserSettingsURL = baseUrl + "/webservices/UserSettings.asmx/GetUserSetting"
    getSideMenusURL = baseUrl + "/webservices/config.asmx/GetSideMenus"
    getGridColumnsForAdvSearchURL = baseUrl + "/webservices/config.asmx/GetGridColumnsForAdvSearch"
    getGridColumnsURL = baseUrl + "/webservices/Config.asmx/GetGridColumns"
    getAuditDataURL = baseUrl + "/webservices/Audit.asmx/GetAuditData"
    getItemsURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetItems"
    getPriceListsURL = baseUrl + "/webservices/PriceListService.asmx/GetPriceLists"
    getInvoiceItemPriceURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetInvoiceItemPrice"
    getPriceListItemURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetPriceListItem"
    getNewItemForQuoteURL = baseUrl + "/webservices/QuoteService.asmx/GetNewItemForQuote"
    getDriversURL = baseUrl + "/webservices/Drivers.asmx/GetAllDrivers"
    getEquipmentURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetEquipment"
    getEmployeesURL = baseUrl + "/webservices/Employees.asmx/GetDrivers"
    getSubItemsURL = baseUrl + "/webservices/GeneralAjaxService.asmx/GetSubItems"

    def __init__(self, email=None, password=None):
        self.tcr_auth = auth(email, password)
        self.headers = self.tcr_auth.header
        self.sideMenu = None


    def getSideMenus(self):
        """Gets the Side Menus from TCR & Sets instance variable sideMenu"""
        request = requests.post(self.getSideMenusURL, headers=self.headers).json()
        self.sideMenus = request["d"]
        return request["d"]


    def manualGetGridData(self, PostDict, includeCount=False):
        """
        Gets the Grid Data from TCR  
            - PostDict is required
            - includeCount is optional

        Parameters
        ----------
            PostDict : dict
                The Dictionary Of Data To Post.
            includeCount : bool
                If True, Includes the Count of the returned Grid Data
        """
        response = requests.post(self.getGridDataURL, headers=self.headers, data=PostDict).json()
        resultjson = json.loads(response["d"]["Result"])
        count = resultjson["RecordCount"]
        data = resultjson["Data"]
        if includeCount:
            return {"count": count, "data": data}
        else:
            return data


    def getGrid(self, grid):
        """
        Gets The Full JSON Response from TCR with Grid Name  
            - Uses GetGridByID url if grid is an int
            - Uses GetGrid url if grid is a string
        
        Parameters:
        ----------
            grid : str

        Returns:
        -------
            dict : contains the grid info
        """
        if isinstance(grid, int):  # ? Uses "getGridByIDURL" if grid is an int
            response = requests.post(self.getGridByIDURL, headers=self.headers, json={"gridID": grid}).json()
            return response["d"]
        if isinstance(grid, str):  # ? Uses "getGridURL" if grid is a str
            response = requests.post(self.getGridURL, headers=self.headers, json={"gridName": grid}).json()
            return response["d"]
        else:
            raise ValueError("Grid must be a string or int")


    def gridNameID(self, grid):
        """
        If Given the GridName, Returns the GridID
        If Given the GridID, Returns the GridName
        """
        gridInfo = self.getGrid(grid)
        if isinstance(grid, str):
            return gridInfo["GridID"]
        if isinstance(grid, int):
            return gridInfo["GridName"]

    def getGridColumns(self, grid):
        """
        Gets the Grid Columns from TCR
        """
        if isinstance(grid, str):
            grid = self.gridNameID(grid)
        response = requests.post(self.getGridColumnsURL, headers=self.headers, json={"gridID": grid}).json()
        return response["d"]


    def getGridDataFields(self, grid=None, columns=None):
        """
        Gets The Required Field Attributes from the Grid Name

        Parameters::
        ----------
            grid -- grid name or grid id : str or int
            columns -- list of columns to search : list of dict


        Returns::
        -------
            list[str] | Data Fields
        """
        if grid is not None:
            grid = self.getGrid(grid)
        if columns is not None:
            grid = columns
        else:
            raise ValueError("Grid or Columns are required to get the Data Fields")
        return [field["DataField"] for field in grid["Columns"]]


    def getFilterFields(self, grid):
        """
        Gets The Default Filter Fields from TCR
            > GridID or GridName is required
        """
        getGrid = self.getGrid(grid)
        filterField = getGrid["FilterFields"].split(";")
        # defaultOperator = getGrid["DefaultSelect"]  # ? Not sure if this is the default operator or not...
        return filterField


    def getGridQuickSearchFields(self, grid=None, columns=None):
        """
        Gets The QuickSearch Fields either grid or columns  

        Parameters::
        ----------
            grid -- grid name or grid id : str or int  
            columns -- list of columns to search : list of dict

        Returns::
        -------
            list[str] | Quick Search Fields
        """
        if grid is not None:
            gridColumns = self.getGrid(grid)["Columns"]
        elif columns is not None:
            gridColumns = columns
        quickSearchFields = [field["DataField"] for field in gridColumns if field["QuickSearch"] is True]
        return quickSearchFields


    def getUserSettings(self, settingName):
        """
        Gets The User Settings from TCR
            > SettingName is required
        """
        response = requests.post(self.getUserSettingsURL, headers=self.headers, json={"settingName": settingName}).json()
        if isinstance(response["d"], str):
            return json.loads(response["d"])
        else:
            return response["d"]


    def getGridSettings(self, grid):
        """
        Gets The Grid Settings from TCR
            > GridID or GridName is required
        """
        if isinstance(grid, str):
            grid = self.gridNameID(grid)
        return self.getUserSettings(f"Yagna.Grid.{grid}")


    def getGridSortSettings(self, grid):
        """
        Gets the Grid Sort Settings if they Exist.
            > GridID or GridName is required
        """
        response = self.getGridSettings(grid)
        if response is None:
            getGridInfo = self.getGrid(grid)
            if getGridInfo["DefaultSortColumn"] is not None:
                return Sort(Attribute=getGridInfo["DefaultSortColumn"], Order=getGridInfo["DefaultSortOrder"]).list()
            else:
                return []
        if response["SortCol"]:
            return Sort(Attribute=response["SortCol"], Order=SortDir.fromSettings(response["SortDir"])).list()
        else:
            return []


    def getColumnsForAdvSearch(self, grid):
        """
        Gets the Columns for Advanced Search from TCR
        
        Parameters::
        ----------
            grid -- grid name or grid id : str or int

        Returns::
        -------
            list[dict] | List of Dictionaries Containing Column Info
        """
        gridName = self.getGrid(grid)["GridName"]
        payload = {"gridName": gridName}
        response = requests.post(self.getGridColumnsForAdvSearchURL, headers=self.headers, json=payload).json()
        return response["d"]


    def getGridData(self, Grid, FilterConditions=None,
                    StartIndex=1, RecordCount=250,
                    QuickSearch=None, IncludeCount=True,
                    ):
        """
        Gets The Grid Data from TCR

        Parameters::
        ----------
            Grid : str or int
            FilterConditions : dict
            StartIndex : int (default=1)
            RecordCount : int (default=250)
            QuickSearch : str (default=None)
            IncludeCount : bool (default=True)

        Returns::
        -------
            dict : contains the grid data and count
        """
        getGridInfo = self.getGrid(Grid)
        gridID = getGridInfo["GridID"]
        attributeFields = [field["DataField"] for field in getGridInfo["Columns"] ]
        dateField = [field["DataField"] for field in getGridInfo["Columns"] if field["DataType"] == 4]

        if FilterConditions is None:
            FilterConditions = Filter()
        elif isinstance(FilterConditions, dict):
            FilterConditions = Filter(**FilterConditions)
        
        if QuickSearch is not None:
            quickSearchFields = [field["DataField"] for field in getGridInfo["Columns"] if field["QuickSearch"] is True]
            FilterConditions.add_advanced_filter(Filter().add_quick_search_filter(QuickSearch, quickSearchFields))

        if isinstance(FilterConditions, Filter):
            FilterConditions = FilterConditions.dict()
            pass

        requestData = GetGridDateRootModel(
            query=GetGridDataModel(
                GridID=getGridInfo["GridID"],
                RecordCount=RecordCount,
                Filter=FilterConditions,
                StartIndex=StartIndex,
                Attributes=attributeFields,
                Sort=self.getGridSortSettings(gridID),
                CustomSort=None,
            )
        ).json()

        response = requests.post(self.getGridDataURL, headers=self.headers, data=requestData).json()
        resultjson = json.loads(response["d"]["Result"])
        for field in dateField:
            for row in resultjson["Data"]:
                if row[field] is not None:
                    splitDot = row[field].split(".")
                    row[field] = dt.datetime.strptime(splitDot[0], "%Y-%m-%dT%H:%M:%S")
        if IncludeCount is True:
            return {"count": resultjson["RecordCount"], "data": resultjson["Data"]}
        else:
            return resultjson["Data"]


    def getTicket(self, ticketID: int):
        """
        Gets the Ticket Details from TCR

        Parameters::
        ----------
            ticketID -- Ticket ID : int

        Returns::
        -------
            dict -- Ticket Details
        """
        ticketID = {"ticketID": ticketID}
        response = requests.post(self.getTicketURL, headers=self.headers, json=ticketID).json()
        responseD = response["d"]
        dateKeys = ["TicketDate", "DelPUDate", "DateCreated", "DateUpdated", "PortalDateCreated",
                    "VoidDate", "DateSigned", "DateSigned_Adjusted", "DriverCompletedDate",
                    "DateFinalEdited", "EstimatedStartTime", "EstimatedEndTime"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD


    def getJob(self, jobID: int):
        """
        Gets the Job Details from TCR

        Parameters::
        ----------
            jobID -- Job ID : int

        Returns::
        -------
            dict -- Job Details
        """
        jobID = {"JobID": jobID}
        response = requests.post(self.getJobURL, headers=self.headers, json=jobID).json()
        responseD = response["d"]
        dateKeys = [
            "DateOpened", "PermitExpiration", "CloseDate", "EstCloseDate", "ContractStart",
            "ContractExpiration", "FirstBillingDate", "LastBillingDate", "DateCreated",
            "DateUpdated", "PortalDateCreated", "PreliminaryNoticeSent", "PreliminaryNoticeReturn",
            "EstStartDate", "FirstTicket", "LastTicket", "LastInvoiceDate"
        ]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD


    def getCustomer(self, custID: int):
        """
        Gets the Customer Details from TCR

        Parameters::
        ----------
            custID -- Customer ID : int

        Returns::
        -------
            dict -- Customer Details
        """
        custID = {"custID": custID}
        response = requests.post(self.getCustomerURL, headers=self.headers, json=custID).json()
        dateKeys = ["DateOpened", "DateCreated", "DateUpdated", "InsurCertExpiration"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD


    def getItems(self):
        """
        Gets all the Items from TCR

        Returns::
        -------
            dict -- Items
        """
        response = requests.post(self.getItemsURL, headers=self.headers).json()
        return response["d"]

    def getPriceLists(self):
        """
        Gets all the Price Lists from TCR

        Returns::
        -------
            dict -- Price Lists
        """
        dateKeys = ["DateCreated", "DateUpdated"]
        response = requests.post(self.getPriceListsURL, headers=self.headers).json()
        for item in response["d"]:
            for key in dateKeys:
                if item[key] is not None:
                    item[key] = millisecond_stamp_to_datetime(item[key])
        return response["d"]

    def getPriceListItem(self, itemID: int, priceID: int):
        """
        Gets the Price List Item Details from TCR

        Parameters::
        ----------
            itemID -- Item ID : int
            priceID -- Price List ID : int

        Returns::
        -------
            dict -- Price List Item Details
        """
        priceListItem = {"itemID": itemID, "priceID": priceID}
        response = requests.post(self.getPriceListItemURL, headers=self.headers, json=priceListItem).json()
        dateKeys = ["DateCreated", "DateUpdated"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
        return responseD

    def getInvoiceItemPrice(self, invoiceID, jobID, itemID):
        """
        Gets the Invoice Item Price from TCR

        Parameters::
        ----------
            invoiceID -- Invoice ID : int
            jobID -- Job ID : int
            itemID -- Item ID : int

        Returns::
        -------
            dict -- Invoice Item Price
        """
        dateKeys = ["DateCreated", "DateUpdated"]
        postData = {"invoiceID": invoiceID, "jobID": jobID, "itemID": itemID}
        response = requests.post(self.getInvoiceItemPriceURL, headers=self.headers, json=postData).json()
        for key in dateKeys:
            if response["d"][key] is not None:
                response["d"][key] = millisecond_stamp_to_datetime(response["d"][key])
        return response["d"]

    def getNewItemForQuote(self, quoteID, itemID):
        """
        Gets the New Item for Quote from TCR

        Parameters::
        ----------
            quoteID -- Quote ID : int
            itemID -- Item ID : int

        Returns::
        -------
            dict -- New Item for Quote
        """
        dateKeys = ["DateCreated", "DateUpdated"]
        postData = {"quoteID": quoteID, "itemID": itemID}
        response = requests.post(self.getNewItemForQuoteURL, headers=self.headers, json=postData).json()
        for key in dateKeys:
            if response["d"][key] is not None:
                response["d"][key] = millisecond_stamp_to_datetime(response["d"][key])
        return response["d"]

    def getCompany(self):
        """
        Gets Your Company Details from TCR

        Returns::
        -------
            dict -- Company Details
        """
        response = requests.post(self.getCompanyURL, headers=self.headers).json()
        responseD = response["d"]
        responseD.pop("LogoImage")
        return responseD

    def getDrivers(self):
        """
        Gets All Drivers From TCR
        
        Returns::
        -------
            list - All Drivers
        """
        response = requests.post(self.getDriversURL, headers=self.headers).json()
        dateKeys = ["LocationDate", "DateCreated", "DateUpdated"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD

    def getEmployees(self):
        """
        Gets All Employees From TCR
        
        Returns::
        -------
            list - All Employees
        """
        response = requests.post(self.getEmployeesURL, headers=self.headers).json()
        dateKeys = ["DateCreated", "DateUpdated"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD

    def getEquipment(self, EquipmentID):
        """
        Gets All Equipment From TCR

        Parameters::
        ----------
            EquipmentID -- Equipment ID : int
        
        Returns::
        -------
            list - All Equipment
        """
        requestData = {"equipID": EquipmentID}
        response = requests.post(self.getEquipmentURL, headers=self.headers, json=requestData).json()
        dateKeys = ["DateCreated", "DateInService", "DateUpdated", "DateOutofService"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD

    def getSubItems(self):
        """
        Gets All Sub Items From TCR
        
        Returns::
        -------
            list - All Sub Items
        """
        response = requests.post(self.getSubItemsURL, headers=self.headers).json()
        dateKeys = ["DateCreated", "DateUpdated"]
        responseD = response["d"]
        for key in dateKeys:
            if responseD[key] is not None:
                responseD[key] = millisecond_stamp_to_datetime(responseD[key])
            if responseD["OriginalRecordData"][key] is not None:
                responseD["OriginalRecordData"][key] = millisecond_stamp_to_datetime(responseD["OriginalRecordData"][key])
        return responseD



import math
import datetime as dt

def millisecond_stamp_to_datetime(milliseconds: int):  # ? As much as i would like to move this to datetime_utils, id rather handlers be independent
    """
    Convert milliseconds to datetime object removing the "/Date(...)/"
    """
    seconds = int("".join([x for x in milliseconds if x.isdigit()])) / 1000  # Remove the "/Date(...)/" and convert to seconds
    return dt.datetime.fromtimestamp(seconds) + dt.timedelta(hours=1)  # Add 1 Hour Since TCR is off by 1 hour

def datetime_to_string(date: dt.datetime):
    """
    ### Converts a datetime object to a string

        Parameters:
            date: The datetime object to convert

        Returns:
            The datetime object as a string
    """
    return date.strftime("%Y-%m-%dT%H:%M:%S.%f")

def page_count(recordCount, pageSize):
    """
    ### Finds the number of pages needed to display all records.

        Parameters:
            recordCount: The total number of records
            pageSize: The amount of records to load on each query

        Returns:
            The number of pages needed to display all records as an integer
    """
    return math.ceil(recordCount / pageSize)


def start_index(pageNumber, pageSize):
    """
    ### Function to find the start index of each page.

        Parameters:
            pageNumber: The page number to find the start index of
            pageSize: The amount of records to load on each query

        Returns:
            The start index of the page as an integer
    """
    return ((pageNumber - 1) * pageSize) + 1



from enum import Enum

class SortDir(Enum):
    ASC = 0
    DESC = 1

    def fromSettings(settings: str):
        """Convert the settings value to a SortDir enum value"""
        if settings == "sort-asc":
            return SortDir.ASC.value
        elif settings == "sort-desc":
            return SortDir.DESC.value
        else:
            return SortDir.ASC.value



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tcrpy",
    version="0.2.5",
    author="Mitch V Aureli",
    author_email="mitch@logic3ii.com",
    description="A Python Library for Interacting with the TCR Platform API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mitchellaha/tcrpy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    install_requires=[
        'beautifulsoup4==4.11.1',
        'pydantic==1.9.0',
        'python-dotenv==0.20.0',
        'requests==2.27.1'
    ],
)


import os
from dotenv import load_dotenv
import requests
from bs4 import BeautifulSoup


class auth:
    loginURL = "http://apps.tcrsoftware.com/tcr_2/login.aspx"
    def __init__(self, email=None, password=None):
        load_dotenv()
        self.email = email or os.getenv("email")
        self.password = password or os.getenv("password")
        self.header = None
        self.cookies = None
        self.expire = None
        self.set_header()


    def get_cookies(self):
        """
        Returns TCR cookies[0] and expiration[1]
        """
        with requests.Session() as s:
            s.headers.update(
                {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36'}
            )
            loginPageContent = s.get(self.loginURL)
            soup = BeautifulSoup(loginPageContent.content, 'html.parser') # Rip the form data from the page with BS4.

            loginPostHeaders = {
                "X-MicrosoftAjax": "Delta=true",
                "X-Requested-With": "XMLHttpRequest",
                "Referer": "https://apps.tcrsoftware.com/tcr_2/login.aspx"
            }

            loginData = {
                "ScriptManager1": "UpdatePanel1|BtnSubmit",
                "__LASTFOCUS": "",
                "__EVENTTARGET": "",
                "__EVENTARGUMENT": "",
                "__VIEWSTATE": str(soup.select_one("#__VIEWSTATE")["value"]),
                "__VIEWSTATEGENERATOR": str(soup.select_one("#__VIEWSTATEGENERATOR")["value"]),
                "__EVENTVALIDATION": str(soup.select_one("#__EVENTVALIDATION")["value"]),
                "Email": str(self.email),
                "Password": str(self.password),
                "Remember": "on",
                "Email2": "",
                "_ASYCNPOST": True,
                "BtnSubmit": "Login",
            }

            s.post(self.loginURL, data=loginData, headers=loginPostHeaders)

            cookies = s.cookies.get_dict()
            expire = {}
            for cookie in s.cookies:
                if cookie.name == "Email":
                    emailExpires = cookie.expires  # ?Todo: Convert from epoch to datetime
                    expire["Email"] = emailExpires
                if cookie.name == "TCRAuth":
                    TCRAuthExpires = cookie.expires  # ?Todo: Convert from epoch to datetime
                    expire["TCRAuth"] = TCRAuthExpires
        return cookies, expire


    def set_header(self):
        cookies, expire = self.get_cookies()
        headers = {
            'DNT': '1',
            'X-Requested-With': 'XMLHttpRequest',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36',
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json'
        }
        cookie = "Email={}; ASP.NET_SessionId={}; TCRAuth={}".format(
            cookies["Email"], cookies["ASP.NET_SessionId"], cookies["TCRAuth"])
        headers["Cookie"] = cookie
        self.header = headers
        self.cookies = cookies
        self.expire = expire
        return headers



from typing import Any, List

from pydantic import BaseModel, conint


class GetGridDataModel(BaseModel):
    GridID: int
    RecordCount: conint(ge=50) = 50
    Filter: dict = {}
    StartIndex: conint(ge=1) = 1
    Attributes: List[str]
    Sort: list = []
    CustomSort: Any = None

class GetGridDateRootModel(BaseModel):
    query: GetGridDataModel


class Filter:
    """
    Used to create filter conditions for the getGrid API Call.
    """
    def __init__(self, Attribute: str = None, Operator: int = None, Values: list = None, **kwargs):
        self.Conditions = []

        if Attribute is not None and Operator is not None and Values is not None:
            self.add_condition(Attribute, Operator, Values)

        if "Conditions" in kwargs.keys():
            self.Conditions = kwargs["Conditions"]
        if "Filter" in kwargs.keys():
            self.Filter = kwargs["Filter"]
        if "GroupOperator" in kwargs.keys():
            self.GroupOperator = kwargs["GroupOperator"]


    def add_condition(self, Attribute: str, Operator: int, Values: list):
        """
        Adds a condition to the filter.
        
        params
        --------
        attribute: str
            The attribute to filter on.
                ex. "OrderType"
        value: list
            The value to filter on.
                ex. ["Name"] or ["Date", "Date"]
        operator: int
            The operator to use.
                ex. <SEE OPERATOR DOCUMENTATION>
        """
        self.Conditions.append({
            "Attribute": Attribute,
            "Operator": Operator,
            "Values": Values
        })
        return self

    def add_advanced_filter(self, filter):
        """
        Adds a Filter Class inside this Filter Class.
        
        params
        --------
        filter: Filter Object or dict
            The filter to add.
        """
        if isinstance(filter, Filter):
            self.Filter = filter.dict()
        elif isinstance(filter, dict) and "Conditions" in filter.keys():
            self.Filter = filter
        else:
            raise TypeError("Filter must be of type Filter or dict.")
        return self

    def set_group_operator(self, Operator: int):
        """
        Adds a group operator to the filter.
        Group Operators are Only Used When Adding Quick Search Filters.
            
        params
        --------
        operator: int
            The operator to use.
        """
        self.GroupOperator = Operator
        return self

    def add_quick_search_filter(self, SearchQuery, ListOfAttributes, GroupOperator = 2):
        """
        Adds Many Search Conditions to the Filter.

        params
        --------
        SearchQuery: str
            The search query to use.
        ListOfAttributes: list
            The list of attributes to search on.
        GroupOperator: int (optional)
            - 1 = AND
            - 2 = OR (Default)

        Example:
        --------
            >>> searchFilterEx = Filter()
            >>> searchFilterEx.AddSearchFilter("MyName", ["Drivers", "JobLocation", "CustomerName"])
            >>> searchFilterEx.dict()
            {
                "Conditions": [
                    {
                        "Attribute": "Drivers",
                        "Operator": 10,
                        "Values": ["MyName"]
                    },
                    {
                        "Attribute": "JobLocation",
                        "Operator": 10,
                        "Values": ["MyName"]
                    },
                    {
                        "Attribute": "CustomerName",
                        "Operator": 10,
                        "Values": ["MyName"]
                    }
                ],
                "GroupOperator": 2
            }
        """
        self.GroupOperator = GroupOperator
        for attribute in ListOfAttributes:
            self.add_condition(attribute, 10, [SearchQuery])
        return self

    def dict(self):
        """
        Returns the filter as a dictionary.
        """
        return self.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return self.__repr__()



class Sort:  # TODO: Does this Need to be Redone to add multiple sorts??
    """
    Used to create sort conditions for the getGrid API Call.

    Example:
    --------
        >>> gridSort = Sort(Attribute="OrderType", Order=1)
        >>> gridSort.list()
        [
            {
                "Attribute": "OrderType",
                "Order": 1
            }
        ]
    """
    def __init__(self, Attribute: str, Order: int):
        self.Attribute = Attribute
        self.Order = Order

    def dict(self):
        """
        Returns the sort as a dictionary.
        """
        return self.__dict__

    def list(self):
        """
        Returns the sort as a list.
        """
        return [self.__dict__]

    def __repr__(self):
        return str(list)

    def __str__(self):
        return self.__repr__()
